@@ -43,18 +43,8 @@ using testing::SizeIs;
 
 namespace
 {
-constexpr auto NODE_SUCCESS =
-  rclcpp_lifecycle::node_interfaces::LifecycleNodeInterface::CallbackReturn::SUCCESS;
-constexpr auto NODE_ERROR =
-  rclcpp_lifecycle::node_interfaces::LifecycleNodeInterface::CallbackReturn::ERROR;
-
-rclcpp::WaitResultKind wait_for(rclcpp::SubscriptionBase::SharedPtr subscription)
-{
-  rclcpp::WaitSet wait_set;
-  wait_set.add_subscription(subscription);
-  const auto timeout = std::chrono::seconds(10);
-  return wait_set.wait(timeout).kind();
-}
+constexpr auto NODE_SUCCESS = controller_interface::CallbackReturn::SUCCESS;
+constexpr auto NODE_ERROR = controller_interface::CallbackReturn::ERROR;
 }  // namespace
 
 void JointStateBroadcasterTest::SetUpTestCase() { rclcpp::init(0, nullptr); }
@@ -158,19 +148,9 @@ void JointStateBroadcasterTest::assign_state_interfaces(
 
 TEST_F(JointStateBroadcasterTest, ConfigureErrorTest)
 {
-  // joint state not initialized yet
-  ASSERT_TRUE(state_broadcaster_->joint_state_msg_.name.empty());
-  ASSERT_TRUE(state_broadcaster_->joint_state_msg_.position.empty());
-  ASSERT_TRUE(state_broadcaster_->joint_state_msg_.velocity.empty());
-  ASSERT_TRUE(state_broadcaster_->joint_state_msg_.effort.empty());
-
-  // dynamic joint state not initialized yet
-  ASSERT_TRUE(state_broadcaster_->dynamic_joint_state_msg_.joint_names.empty());
-  ASSERT_TRUE(state_broadcaster_->dynamic_joint_state_msg_.interface_values.empty());
-
   // publishers not initialized yet
-  ASSERT_FALSE(state_broadcaster_->joint_state_publisher_);
-  ASSERT_FALSE(state_broadcaster_->dynamic_joint_state_publisher_);
+  ASSERT_FALSE(state_broadcaster_->realtime_joint_state_publisher_);
+  ASSERT_FALSE(state_broadcaster_->realtime_dynamic_joint_state_publisher_);
 
   // configure failed
   ASSERT_THROW(state_broadcaster_->on_configure(rclcpp_lifecycle::State()), std::exception);
@@ -178,33 +158,13 @@ TEST_F(JointStateBroadcasterTest, ConfigureErrorTest)
   SetUpStateBroadcaster();
   // check state remains unchanged
 
-  // joint state still not initialized
-  ASSERT_TRUE(state_broadcaster_->joint_state_msg_.name.empty());
-  ASSERT_TRUE(state_broadcaster_->joint_state_msg_.position.empty());
-  ASSERT_TRUE(state_broadcaster_->joint_state_msg_.velocity.empty());
-  ASSERT_TRUE(state_broadcaster_->joint_state_msg_.effort.empty());
-
-  // dynamic joint state still not initialized
-  ASSERT_TRUE(state_broadcaster_->dynamic_joint_state_msg_.joint_names.empty());
-  ASSERT_TRUE(state_broadcaster_->dynamic_joint_state_msg_.interface_values.empty());
-
   // publishers still not initialized
-  ASSERT_FALSE(state_broadcaster_->joint_state_publisher_);
-  ASSERT_FALSE(state_broadcaster_->dynamic_joint_state_publisher_);
+  ASSERT_FALSE(state_broadcaster_->realtime_joint_state_publisher_);
+  ASSERT_FALSE(state_broadcaster_->realtime_dynamic_joint_state_publisher_);
 }
 
 TEST_F(JointStateBroadcasterTest, ActivateTest)
 {
-  // joint state not initialized yet
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.name, IsEmpty());
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.position, IsEmpty());
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.velocity, IsEmpty());
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.effort, IsEmpty());
-
-  // dynamic joint state not initialized yet
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.joint_names, IsEmpty());
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.interface_values, IsEmpty());
-
   // publishers not initialized yet
   ASSERT_FALSE(state_broadcaster_->joint_state_publisher_);
   ASSERT_FALSE(state_broadcaster_->dynamic_joint_state_publisher_);
@@ -217,30 +177,32 @@ TEST_F(JointStateBroadcasterTest, ActivateTest)
 
   const size_t NUM_JOINTS = joint_names_.size();
 
+  // publishers initialized
+  ASSERT_TRUE(state_broadcaster_->realtime_joint_state_publisher_);
+  ASSERT_TRUE(state_broadcaster_->realtime_dynamic_joint_state_publisher_);
+
   // joint state initialized
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.name, ElementsAreArray(joint_names_));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.position, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.velocity, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.effort, SizeIs(NUM_JOINTS));
+  const auto & joint_state_msg = state_broadcaster_->realtime_joint_state_publisher_->msg_;
+  ASSERT_THAT(joint_state_msg.name, ElementsAreArray(joint_names_));
+  ASSERT_THAT(joint_state_msg.position, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.velocity, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.effort, SizeIs(NUM_JOINTS));
 
   // dynamic joint state initialized
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.joint_names, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.interface_values, SizeIs(NUM_JOINTS));
+  const auto & dynamic_joint_state_msg =
+    state_broadcaster_->realtime_dynamic_joint_state_publisher_->msg_;
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.interface_values, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, ElementsAreArray(joint_names_));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.joint_names, ElementsAreArray(joint_names_));
-  ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[0].interface_names,
+    dynamic_joint_state_msg.interface_values[0].interface_names,
     ElementsAreArray(interface_names_));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[1].interface_names,
+    dynamic_joint_state_msg.interface_values[1].interface_names,
     ElementsAreArray(interface_names_));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[2].interface_names,
+    dynamic_joint_state_msg.interface_values[2].interface_names,
     ElementsAreArray(interface_names_));
-
-  // publishers initialized
-  ASSERT_TRUE(state_broadcaster_->joint_state_publisher_);
-  ASSERT_TRUE(state_broadcaster_->dynamic_joint_state_publisher_);
 }
 
 TEST_F(JointStateBroadcasterTest, ActivateTestWithoutJointsParameter)
@@ -256,30 +218,32 @@ TEST_F(JointStateBroadcasterTest, ActivateTestWithoutJointsParameter)
 
   const size_t NUM_JOINTS = joint_names_.size();
 
+  // publishers initialized
+  ASSERT_TRUE(state_broadcaster_->realtime_joint_state_publisher_);
+  ASSERT_TRUE(state_broadcaster_->realtime_dynamic_joint_state_publisher_);
+
   // joint state initialized
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.name, ElementsAreArray(joint_names_));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.position, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.velocity, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.effort, SizeIs(NUM_JOINTS));
+  const auto & joint_state_msg = state_broadcaster_->realtime_joint_state_publisher_->msg_;
+  ASSERT_THAT(joint_state_msg.name, ElementsAreArray(joint_names_));
+  ASSERT_THAT(joint_state_msg.position, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.velocity, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.effort, SizeIs(NUM_JOINTS));
 
   // dynamic joint state initialized
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.joint_names, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.interface_values, SizeIs(NUM_JOINTS));
+  const auto & dynamic_joint_state_msg =
+    state_broadcaster_->realtime_dynamic_joint_state_publisher_->msg_;
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.interface_values, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, ElementsAreArray(joint_names_));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.joint_names, ElementsAreArray(joint_names_));
-  ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[0].interface_names,
+    dynamic_joint_state_msg.interface_values[0].interface_names,
     ElementsAreArray(interface_names_));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[1].interface_names,
+    dynamic_joint_state_msg.interface_values[1].interface_names,
     ElementsAreArray(interface_names_));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[2].interface_names,
+    dynamic_joint_state_msg.interface_values[2].interface_names,
     ElementsAreArray(interface_names_));
-
-  // publishers initialized
-  ASSERT_TRUE(state_broadcaster_->joint_state_publisher_);
-  ASSERT_TRUE(state_broadcaster_->dynamic_joint_state_publisher_);
 }
 
 TEST_F(JointStateBroadcasterTest, ActivateTestWithoutInterfacesParameter)
@@ -295,30 +259,32 @@ TEST_F(JointStateBroadcasterTest, ActivateTestWithoutInterfacesParameter)
 
   const size_t NUM_JOINTS = joint_names_.size();
 
+  // publishers initialized
+  ASSERT_TRUE(state_broadcaster_->realtime_joint_state_publisher_);
+  ASSERT_TRUE(state_broadcaster_->realtime_dynamic_joint_state_publisher_);
+
   // joint state initialized
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.name, ElementsAreArray(joint_names_));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.position, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.velocity, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.effort, SizeIs(NUM_JOINTS));
+  const auto & joint_state_msg = state_broadcaster_->realtime_joint_state_publisher_->msg_;
+  ASSERT_THAT(joint_state_msg.name, ElementsAreArray(joint_names_));
+  ASSERT_THAT(joint_state_msg.position, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.velocity, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.effort, SizeIs(NUM_JOINTS));
 
   // dynamic joint state initialized
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.joint_names, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.interface_values, SizeIs(NUM_JOINTS));
+  const auto & dynamic_joint_state_msg =
+    state_broadcaster_->realtime_dynamic_joint_state_publisher_->msg_;
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.interface_values, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, ElementsAreArray(joint_names_));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.joint_names, ElementsAreArray(joint_names_));
-  ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[0].interface_names,
+    dynamic_joint_state_msg.interface_values[0].interface_names,
     ElementsAreArray(interface_names_));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[1].interface_names,
+    dynamic_joint_state_msg.interface_values[1].interface_names,
     ElementsAreArray(interface_names_));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[2].interface_names,
+    dynamic_joint_state_msg.interface_values[2].interface_names,
     ElementsAreArray(interface_names_));
-
-  // publishers initialized
-  ASSERT_TRUE(state_broadcaster_->joint_state_publisher_);
-  ASSERT_TRUE(state_broadcaster_->dynamic_joint_state_publisher_);
 }
 
 TEST_F(JointStateBroadcasterTest, ActivateTestTwoJointsOneInterface)
@@ -334,35 +300,35 @@ TEST_F(JointStateBroadcasterTest, ActivateTestTwoJointsOneInterface)
 
   const size_t NUM_JOINTS = JOINT_NAMES.size();
 
+  // publishers initialized
+  ASSERT_TRUE(state_broadcaster_->realtime_joint_state_publisher_);
+  ASSERT_TRUE(state_broadcaster_->realtime_dynamic_joint_state_publisher_);
+
   // joint state initialized
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.name, ElementsAreArray(JOINT_NAMES));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.position, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.velocity, SizeIs(NUM_JOINTS));
+  const auto & joint_state_msg = state_broadcaster_->realtime_joint_state_publisher_->msg_;
+  ASSERT_THAT(joint_state_msg.name, ElementsAreArray(JOINT_NAMES));
+  ASSERT_THAT(joint_state_msg.position, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.velocity, SizeIs(NUM_JOINTS));
   for (auto i = 0ul; i < NUM_JOINTS; ++i)
   {
-    ASSERT_TRUE(std::isnan(state_broadcaster_->joint_state_msg_.velocity[i]));
+    ASSERT_TRUE(std::isnan(joint_state_msg.velocity[i]));
   }
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.effort, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.effort, SizeIs(NUM_JOINTS));
   for (auto i = 0ul; i < NUM_JOINTS; ++i)
   {
-    ASSERT_TRUE(std::isnan(state_broadcaster_->joint_state_msg_.effort[i]));
+    ASSERT_TRUE(std::isnan(joint_state_msg.effort[i]));
   }
 
   // dynamic joint state initialized
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.joint_names, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.interface_values, SizeIs(NUM_JOINTS));
+  const auto & dynamic_joint_state_msg =
+    state_broadcaster_->realtime_dynamic_joint_state_publisher_->msg_;
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.interface_values, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, ElementsAreArray(JOINT_NAMES));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.joint_names, ElementsAreArray(JOINT_NAMES));
+    dynamic_joint_state_msg.interface_values[0].interface_names, ElementsAreArray(IF_NAMES));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[0].interface_names,
-    ElementsAreArray(IF_NAMES));
-  ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[1].interface_names,
-    ElementsAreArray(IF_NAMES));
-
-  // publishers initialized
-  ASSERT_TRUE(state_broadcaster_->joint_state_publisher_);
-  ASSERT_TRUE(state_broadcaster_->dynamic_joint_state_publisher_);
+    dynamic_joint_state_msg.interface_values[1].interface_names, ElementsAreArray(IF_NAMES));
 }
 
 TEST_F(JointStateBroadcasterTest, ActivateTestOneJointTwoInterfaces)
@@ -378,28 +344,29 @@ TEST_F(JointStateBroadcasterTest, ActivateTestOneJointTwoInterfaces)
 
   const size_t NUM_JOINTS = JOINT_NAMES.size();
 
+  // publishers initialized
+  ASSERT_TRUE(state_broadcaster_->realtime_joint_state_publisher_);
+  ASSERT_TRUE(state_broadcaster_->realtime_dynamic_joint_state_publisher_);
+
   // joint state initialized
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.name, ElementsAreArray(JOINT_NAMES));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.position, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.velocity, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.effort, SizeIs(NUM_JOINTS));
+  const auto & joint_state_msg = state_broadcaster_->realtime_joint_state_publisher_->msg_;
+  ASSERT_THAT(joint_state_msg.name, ElementsAreArray(JOINT_NAMES));
+  ASSERT_THAT(joint_state_msg.position, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.velocity, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.effort, SizeIs(NUM_JOINTS));
   for (auto i = 0ul; i < NUM_JOINTS; ++i)
   {
-    ASSERT_TRUE(std::isnan(state_broadcaster_->joint_state_msg_.effort[i]));
+    ASSERT_TRUE(std::isnan(joint_state_msg.effort[i]));
   }
 
   // dynamic joint state initialized
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.joint_names, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.interface_values, SizeIs(NUM_JOINTS));
+  const auto & dynamic_joint_state_msg =
+    state_broadcaster_->realtime_dynamic_joint_state_publisher_->msg_;
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.interface_values, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, ElementsAreArray(JOINT_NAMES));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.joint_names, ElementsAreArray(JOINT_NAMES));
-  ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[0].interface_names,
-    ElementsAreArray(IF_NAMES));
-
-  // publishers initialized
-  ASSERT_TRUE(state_broadcaster_->joint_state_publisher_);
-  ASSERT_TRUE(state_broadcaster_->dynamic_joint_state_publisher_);
+    dynamic_joint_state_msg.interface_values[0].interface_names, ElementsAreArray(IF_NAMES));
 }
 
 TEST_F(JointStateBroadcasterTest, ActivateTestTwoJointTwoInterfacesAllMissing)
@@ -445,33 +412,34 @@ TEST_F(JointStateBroadcasterTest, ActivateTestTwoJointTwoInterfacesOneMissing)
 
   const size_t NUM_JOINTS = JOINT_NAMES.size();
 
+  // publishers initialized
+  ASSERT_TRUE(state_broadcaster_->realtime_joint_state_publisher_);
+  ASSERT_TRUE(state_broadcaster_->realtime_dynamic_joint_state_publisher_);
+
   // joint state initialized
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.name, ElementsAreArray(JOINT_NAMES));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.position, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.velocity, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.effort, SizeIs(NUM_JOINTS));
+  const auto & joint_state_msg = state_broadcaster_->realtime_joint_state_publisher_->msg_;
+  ASSERT_THAT(joint_state_msg.name, ElementsAreArray(JOINT_NAMES));
+  ASSERT_THAT(joint_state_msg.position, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.velocity, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.effort, SizeIs(NUM_JOINTS));
   // velocity for joint 1 should be nan because state interface does not exit
-  ASSERT_TRUE(std::isnan(state_broadcaster_->joint_state_msg_.velocity[0]));
+  ASSERT_TRUE(std::isnan(joint_state_msg.velocity[0]));
   for (auto i = 0ul; i < NUM_JOINTS; ++i)
   {
-    ASSERT_TRUE(std::isnan(state_broadcaster_->joint_state_msg_.effort[i]));
+    ASSERT_TRUE(std::isnan(joint_state_msg.effort[i]));
   }
 
   // dynamic joint state initialized
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.joint_names, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.interface_values, SizeIs(NUM_JOINTS));
+  const auto & dynamic_joint_state_msg =
+    state_broadcaster_->realtime_dynamic_joint_state_publisher_->msg_;
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.interface_values, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, ElementsAreArray(JOINT_NAMES));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.joint_names, ElementsAreArray(JOINT_NAMES));
-  ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[0].interface_names,
+    dynamic_joint_state_msg.interface_values[0].interface_names,
     ElementsAreArray({IF_NAMES[0]}));  // joint 1 has only pos interface
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[1].interface_names,
-    ElementsAreArray(IF_NAMES));
-
-  // publishers initialized
-  ASSERT_TRUE(state_broadcaster_->joint_state_publisher_);
-  ASSERT_TRUE(state_broadcaster_->dynamic_joint_state_publisher_);
+    dynamic_joint_state_msg.interface_values[1].interface_names, ElementsAreArray(IF_NAMES));
 }
 
 TEST_F(JointStateBroadcasterTest, TestCustomInterfaceWithoutMapping)
@@ -488,19 +456,20 @@ TEST_F(JointStateBroadcasterTest, TestCustomInterfaceWithoutMapping)
   const size_t NUM_JOINTS = JOINT_NAMES.size();
 
   // joint state initialized
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.name, SizeIs(0));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.position, SizeIs(0));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.velocity, SizeIs(0));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.effort, SizeIs(0));
+  const auto & joint_state_msg = state_broadcaster_->realtime_joint_state_publisher_->msg_;
+  ASSERT_THAT(joint_state_msg.name, SizeIs(0));
+  ASSERT_THAT(joint_state_msg.position, SizeIs(0));
+  ASSERT_THAT(joint_state_msg.velocity, SizeIs(0));
+  ASSERT_THAT(joint_state_msg.effort, SizeIs(0));
 
   // dynamic joint state initialized
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.joint_names, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.interface_values, SizeIs(NUM_JOINTS));
+  const auto & dynamic_joint_state_msg =
+    state_broadcaster_->realtime_dynamic_joint_state_publisher_->msg_;
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.interface_values, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, ElementsAreArray(JOINT_NAMES));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.joint_names, ElementsAreArray(JOINT_NAMES));
-  ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[0].interface_names,
-    ElementsAreArray(IF_NAMES));
+    dynamic_joint_state_msg.interface_values[0].interface_names, ElementsAreArray(IF_NAMES));
 
   // publishers initialized
   ASSERT_TRUE(state_broadcaster_->joint_state_publisher_);
@@ -524,26 +493,28 @@ TEST_F(JointStateBroadcasterTest, TestCustomInterfaceMapping)
   const size_t NUM_JOINTS = JOINT_NAMES.size();
 
   // joint state initialized
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.name, ElementsAreArray(JOINT_NAMES));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.position, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.velocity, SizeIs(NUM_JOINTS));
+  const auto & joint_state_msg = state_broadcaster_->realtime_joint_state_publisher_->msg_;
+  ASSERT_THAT(joint_state_msg.name, ElementsAreArray(JOINT_NAMES));
+  ASSERT_THAT(joint_state_msg.position, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.velocity, SizeIs(NUM_JOINTS));
   for (auto i = 0ul; i < NUM_JOINTS; ++i)
   {
-    ASSERT_TRUE(std::isnan(state_broadcaster_->joint_state_msg_.velocity[i]));
+    ASSERT_TRUE(std::isnan(joint_state_msg.velocity[i]));
   }
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.effort, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.effort, SizeIs(NUM_JOINTS));
   for (auto i = 0ul; i < NUM_JOINTS; ++i)
   {
-    ASSERT_TRUE(std::isnan(state_broadcaster_->joint_state_msg_.effort[i]));
+    ASSERT_TRUE(std::isnan(joint_state_msg.effort[i]));
   }
 
   // dynamic joint state initialized
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.joint_names, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.interface_values, SizeIs(NUM_JOINTS));
+  const auto & dynamic_joint_state_msg =
+    state_broadcaster_->realtime_dynamic_joint_state_publisher_->msg_;
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.interface_values, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, ElementsAreArray(JOINT_NAMES));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.joint_names, ElementsAreArray(JOINT_NAMES));
-  ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[0].interface_names,
+    dynamic_joint_state_msg.interface_values[0].interface_names,
     ElementsAreArray({HW_IF_POSITION}));  // mapping to this value
 
   // publishers initialized
@@ -572,28 +543,30 @@ TEST_F(JointStateBroadcasterTest, TestCustomInterfaceMappingUpdate)
   const size_t NUM_JOINTS = JOINT_NAMES.size();
 
   // joint state initialized
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.name, ElementsAreArray(JOINT_NAMES));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.position, SizeIs(NUM_JOINTS));
-  ASSERT_EQ(state_broadcaster_->joint_state_msg_.position[0], custom_joint_value_);
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.velocity, SizeIs(NUM_JOINTS));
+  const auto & joint_state_msg = state_broadcaster_->realtime_joint_state_publisher_->msg_;
+  ASSERT_THAT(joint_state_msg.name, ElementsAreArray(JOINT_NAMES));
+  ASSERT_THAT(joint_state_msg.position, SizeIs(NUM_JOINTS));
+  ASSERT_EQ(joint_state_msg.position[0], custom_joint_value_);
+  ASSERT_THAT(joint_state_msg.velocity, SizeIs(NUM_JOINTS));
   for (auto i = 0ul; i < NUM_JOINTS; ++i)
   {
-    ASSERT_TRUE(std::isnan(state_broadcaster_->joint_state_msg_.velocity[i]));
+    ASSERT_TRUE(std::isnan(joint_state_msg.velocity[i]));
   }
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.effort, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.effort, SizeIs(NUM_JOINTS));
   for (auto i = 0ul; i < NUM_JOINTS; ++i)
   {
-    ASSERT_TRUE(std::isnan(state_broadcaster_->joint_state_msg_.effort[i]));
+    ASSERT_TRUE(std::isnan(joint_state_msg.effort[i]));
   }
 
   // dynamic joint state initialized
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.joint_names, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.interface_values, SizeIs(NUM_JOINTS));
+  const auto & dynamic_joint_state_msg =
+    state_broadcaster_->realtime_dynamic_joint_state_publisher_->msg_;
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.interface_values, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, ElementsAreArray(JOINT_NAMES));
   ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.joint_names, ElementsAreArray(JOINT_NAMES));
-  ASSERT_THAT(
-    state_broadcaster_->dynamic_joint_state_msg_.interface_values[0].interface_names,
-    ElementsAreArray({HW_IF_POSITION}));  // mapping to this value
+    dynamic_joint_state_msg.interface_values[0].interface_names,
+    ElementsAreArray({HW_IF_POSITION}));
 
   // publishers initialized
   ASSERT_TRUE(state_broadcaster_->joint_state_publisher_);
@@ -604,8 +577,8 @@ TEST_F(JointStateBroadcasterTest, UpdateTest)
 {
   SetUpStateBroadcaster();
 
-  auto node_state = state_broadcaster_->configure();
-  node_state = state_broadcaster_->activate();
+  auto node_state = state_broadcaster_->get_node()->configure();
+  node_state = state_broadcaster_->get_node()->activate();
   ASSERT_EQ(node_state.id(), lifecycle_msgs::msg::State::PRIMARY_STATE_ACTIVE);
   ASSERT_EQ(
     state_broadcaster_->update(rclcpp::Time(0), rclcpp::Duration::from_seconds(0.01)),
@@ -614,10 +587,10 @@ TEST_F(JointStateBroadcasterTest, UpdateTest)
 
 void JointStateBroadcasterTest::test_published_joint_state_message(const std::string & topic)
 {
-  auto node_state = state_broadcaster_->configure();
+  auto node_state = state_broadcaster_->get_node()->configure();
   ASSERT_EQ(node_state.id(), lifecycle_msgs::msg::State::PRIMARY_STATE_INACTIVE);
 
-  node_state = state_broadcaster_->activate();
+  node_state = state_broadcaster_->get_node()->activate();
   ASSERT_EQ(node_state.id(), lifecycle_msgs::msg::State::PRIMARY_STATE_ACTIVE);
 
   rclcpp::Node test_node("test_node");
@@ -625,12 +598,22 @@ void JointStateBroadcasterTest::test_published_joint_state_message(const std::st
   auto subscription =
     test_node.create_subscription<sensor_msgs::msg::JointState>(topic, 10, subs_callback);
 
-  ASSERT_EQ(
-    state_broadcaster_->update(rclcpp::Time(0), rclcpp::Duration::from_seconds(0.01)),
-    controller_interface::return_type::OK);
-
-  // wait for message to be passed
-  ASSERT_EQ(wait_for(subscription), rclcpp::WaitResultKind::Ready);
+  // call update to publish the test value
+  // since update doesn't guarantee a published message, republish until received
+  int max_sub_check_loop_count = 5;  // max number of tries for pub/sub loop
+  rclcpp::WaitSet wait_set;          // block used to wait on message
+  wait_set.add_subscription(subscription);
+  while (max_sub_check_loop_count--)
+  {
+    state_broadcaster_->update(rclcpp::Time(0), rclcpp::Duration::from_seconds(0.01));
+    // check if message has been received
+    if (wait_set.wait(std::chrono::milliseconds(2)).kind() == rclcpp::WaitResultKind::Ready)
+    {
+      break;
+    }
+  }
+  ASSERT_GE(max_sub_check_loop_count, 0) << "Test was unable to publish a message through "
+                                            "controller/broadcaster update loop";
 
   // take message from subscription
   sensor_msgs::msg::JointState joint_state_msg;
@@ -666,10 +649,10 @@ TEST_F(JointStateBroadcasterTest, JointStatePublishTestLocalTopic)
 void JointStateBroadcasterTest::test_published_dynamic_joint_state_message(
   const std::string & topic)
 {
-  auto node_state = state_broadcaster_->configure();
+  auto node_state = state_broadcaster_->get_node()->configure();
   ASSERT_EQ(node_state.id(), lifecycle_msgs::msg::State::PRIMARY_STATE_INACTIVE);
 
-  node_state = state_broadcaster_->activate();
+  node_state = state_broadcaster_->get_node()->activate();
   ASSERT_EQ(node_state.id(), lifecycle_msgs::msg::State::PRIMARY_STATE_ACTIVE);
 
   rclcpp::Node test_node("test_node");
@@ -677,12 +660,22 @@ void JointStateBroadcasterTest::test_published_dynamic_joint_state_message(
   auto subscription =
     test_node.create_subscription<control_msgs::msg::DynamicJointState>(topic, 10, subs_callback);
 
-  ASSERT_EQ(
-    state_broadcaster_->update(rclcpp::Time(0), rclcpp::Duration::from_seconds(0.01)),
-    controller_interface::return_type::OK);
-
-  // wait for message to be passed
-  ASSERT_EQ(wait_for(subscription), rclcpp::WaitResultKind::Ready);
+  // call update to publish the test value
+  // since update doesn't guarantee a published message, republish until received
+  int max_sub_check_loop_count = 5;  // max number of tries for pub/sub loop
+  rclcpp::WaitSet wait_set;          // block used to wait on message
+  wait_set.add_subscription(subscription);
+  while (max_sub_check_loop_count--)
+  {
+    state_broadcaster_->update(rclcpp::Time(0), rclcpp::Duration::from_seconds(0.01));
+    // check if message has been received
+    if (wait_set.wait(std::chrono::milliseconds(2)).kind() == rclcpp::WaitResultKind::Ready)
+    {
+      break;
+    }
+  }
+  ASSERT_GE(max_sub_check_loop_count, 0) << "Test was unable to publish a message through "
+                                            "controller/broadcaster update loop";
 
   // take message from subscription
   control_msgs::msg::DynamicJointState dynamic_joint_state_msg;
@@ -690,7 +683,7 @@ void JointStateBroadcasterTest::test_published_dynamic_joint_state_message(
   ASSERT_TRUE(subscription->take(dynamic_joint_state_msg, msg_info));
 
   const size_t NUM_JOINTS = 3;
-  const auto INTERFACE_NAMES = {HW_IF_POSITION, HW_IF_VELOCITY, HW_IF_EFFORT};
+  const std::vector<std::string> INTERFACE_NAMES = {HW_IF_POSITION, HW_IF_VELOCITY, HW_IF_EFFORT};
   ASSERT_THAT(dynamic_joint_state_msg.joint_names, SizeIs(NUM_JOINTS));
   // the order in the message may be different
   // we only check that all values in this test are present in the message
@@ -728,19 +721,9 @@ TEST_F(JointStateBroadcasterTest, DynamicJointStatePublishTestLocalTopic)
 
 TEST_F(JointStateBroadcasterTest, ExtraJointStatePublishTest)
 {
-  // joint state not initialized yet
-  ASSERT_TRUE(state_broadcaster_->joint_state_msg_.name.empty());
-  ASSERT_TRUE(state_broadcaster_->joint_state_msg_.position.empty());
-  ASSERT_TRUE(state_broadcaster_->joint_state_msg_.velocity.empty());
-  ASSERT_TRUE(state_broadcaster_->joint_state_msg_.effort.empty());
-
-  // dynamic joint state not initialized yet
-  ASSERT_TRUE(state_broadcaster_->dynamic_joint_state_msg_.joint_names.empty());
-  ASSERT_TRUE(state_broadcaster_->dynamic_joint_state_msg_.interface_values.empty());
-
   // publishers not initialized yet
-  ASSERT_FALSE(state_broadcaster_->joint_state_publisher_);
-  ASSERT_FALSE(state_broadcaster_->dynamic_joint_state_publisher_);
+  ASSERT_FALSE(state_broadcaster_->realtime_joint_state_publisher_);
+  ASSERT_FALSE(state_broadcaster_->realtime_dynamic_joint_state_publisher_);
 
   SetUpStateBroadcaster();
 
@@ -758,11 +741,14 @@ TEST_F(JointStateBroadcasterTest, ExtraJointStatePublishTest)
   const size_t NUM_JOINTS = all_joint_names.size();
 
   // joint state initialized
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.name, ElementsAreArray(all_joint_names));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.position, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.velocity, SizeIs(NUM_JOINTS));
-  ASSERT_THAT(state_broadcaster_->joint_state_msg_.effort, SizeIs(NUM_JOINTS));
+  const auto & joint_state_msg = state_broadcaster_->realtime_joint_state_publisher_->msg_;
+  ASSERT_THAT(joint_state_msg.name, ElementsAreArray(all_joint_names));
+  ASSERT_THAT(joint_state_msg.position, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.velocity, SizeIs(NUM_JOINTS));
+  ASSERT_THAT(joint_state_msg.effort, SizeIs(NUM_JOINTS));
 
   // dynamic joint state initialized
-  ASSERT_THAT(state_broadcaster_->dynamic_joint_state_msg_.joint_names, SizeIs(NUM_JOINTS));
+  const auto & dynamic_joint_state_msg =
+    state_broadcaster_->realtime_dynamic_joint_state_publisher_->msg_;
+  ASSERT_THAT(dynamic_joint_state_msg.joint_names, SizeIs(NUM_JOINTS));
 }
